# import pkgutil
import inspect
import os
import sys
import traceback

from birgitta.dataframesource import contextsource
from pyspark.sql.utils import AnalysisException

__all__ = ['exec_code', 'run', 'run_and_exit']


def exec_code(code, globals_dict):
    """Execute the recipe code and handle error conditions.

    Args:
        code (str): The code to be executed.
        globals_dict (dict): globals_dict as used by exec().


    Returns:
       Output of python exec() function.
    """
    e = False
    try:
        ret = exec(code, globals_dict)
    except SyntaxError as err:
        err_desc = err.__class__.__name__
        # detail = err.args[0]
        lineno = err.lineno
        e = err
    except AnalysisException as err:
        err_desc = err.__class__.__name__
        lineno = get_lineno(err)
        e = err
    except NameError as err:
        err_desc = f'{err.__class__.__name__}: {err}'
        lineno = get_lineno(err) - 1
        if lineno < 0:
            lineno = 0
        e = err
    except Exception as err:
        err_desc = err.__class__.__name__
        lineno = get_lineno(err)
        e = err
    if e:
        lines = code.splitlines()
        lenlines = len(lines)
        if lineno < lenlines:
            print("\nRecipe execution",
                  err_desc,
                  "\n",
                  guilty_lines(lines, lineno))
        else:
            print("Recipe execution",
                  err_desc)
        raise e
    return ret


def get_lineno(err):
    # detail = err.args[0]
    cl, exc, tb = sys.exc_info()
    tracebacks = traceback.extract_tb(tb)
    filename = None
    for tback in tracebacks:
        filename = tback[0]
        if filename == '<string>':  # Found exec()'ed string entry
            lineno = tback[1]
            break
    if not filename:
        # Fall back to last line, if filename == '<string>' not found
        lineno = tracebacks[-1][1]
    return lineno


def guilty_lines(lines, lineno):
    """Print the error line, including previous and next lines"""
    # Previous lines
    ret_lines = [""]  # Add spacing
    non_prefix = "         "
    err_prefix = "ERROR -> "
    for i in range(-6, 0):
        prev_lineno = lineno + i
        if prev_lineno >= 0:
            line = lines[prev_lineno]
            if line.startswith('log_transform('):
                continue  # Skip injected log lines
            ret_lines.append(printable_line(non_prefix,
                                            prev_lineno,
                                            line))
    # Error line
    ret_lines.append(printable_line(err_prefix, lineno, lines[lineno]))
    # Next lines
    lenlines = len(lines)
    for i in range(1, 7):
        next_lineno = lineno + i
        if next_lineno < lenlines:
            line = lines[next_lineno]
            if line.startswith('log_transform('):
                continue  # Skip injected log lines
            ret_lines.append(
                printable_line(non_prefix, next_lineno, line))
    ret_lines.append("")  # Add spacing
    return "\n".join(ret_lines)


def printable_line(prefix, lineno, line):
    # FUTURE: Detract script prepend size from lineno, so lineno corresponds
    # to recipe code line no
    # return f"{non_prefix} {prev_lineno}: {lines[prev_lineno]}"
    return f"{prefix}: {line}"


def run(root_mod, recipe, *, dataframe_source=None, replacements=[]):
    """Run a recipe stored in .py file with exec(). The path is relative
    to the path of the mod.

    Args:
        root_mod (module): The root module on which to base the path.
        recipe (str): Relative path to the recipe file from the module dir.
        dataframe_source (DataframeSourceBase subclass): dataframe source
            E.g. LocalSource, Dataiku
        replacements (list): List of text replacements to enable recipe
        debugging. Example on how to limit data amount:

        [
            {
                "old": "dataframe.get(spark_session, ds_foo.name)",
                "new": "dataframe.get(spark_session, ds_foo.name).limit(10)"
            }
        ]

    Returns:
       Output of python exec() function.
    """
    rpath = recipe_path(root_mod, recipe)
    with open(rpath) as f:
        code = prepare_code(f.read(), recipe, replacements)
    if not dataframe_source:
        dataframe_source = contextsource.get()
    globals_dict = {
        'BIRGITTA_DATAFRAMESOURCE': dataframe_source
    }
    return exec_code(code, globals_dict)


def run_and_exit(root_mod, recipe, *, dataframe_source=None, replacements=[]):
    """Run a recipe stored in .py file with exec(). The path is relative
    to the path of the mod. When finished exit(). This is a utility function
    to shortcut a recipe, and leave the rest of the recipe unexecuted.
    This way the recipe can easily be reenabled if further hacking is needed.

    Args:
        root_mod (module): The root module on which to base the path.
        recipe (str): Relative path to the recipe file from the module dir.
        dataframe_source (DataframeSourceBase subclass): dataframe source
            E.g. LocalSource, Dataiku
        replacements (list): List of text replacements to enable recipe
        debugging. Example on how to limit data amount:

        [
            {
                "old": "dataframe.get(spark_session, ds_foo.name)",
                "new": "dataframe.get(spark_session, ds_foo.name).limit(10)"
            }
        ]

    Returns:
       None. Prints output and calls sys.exit().
    """
    ret = run(root_mod,
              recipe,
              dataframe_source=dataframe_source,
              replacements=replacements)
    print(ret)
    rpath = recipe_path(root_mod, recipe)
    print(f"Exit after successfully running recipe: {rpath}")
    sys.exit(0)


def prepare_code(code, recipe, replacements):
    for replacement in replacements:
        code = code.replace(replacement["old"], replacement["new"])
    context_stmts = f"""from birgitta.dataframesource import contextsource
contextsource.set(globals()['BIRGITTA_DATAFRAMESOURCE'])
"""
    completed = f"""
print("=== Recipe {recipe} complete ===")
"""
    return context_stmts + code + completed


def recipe_path(root_mod, recipe):
    mod_path = os.path.dirname(inspect.getfile(root_mod))
    return f"{mod_path}/{recipe}"



"""Helpers and imports for pyspark unit tests
"""
import pytest
from birgitta import context
from birgitta import spark
from birgitta.schema.fixtures import values

# We must import these fixtures explicitly to make them available
# in localtest fixtures at run time.
from birgitta.recipetest.coverage.report import cov_report  # noqa 401
from birgitta.recipetest.coverage.report import cov_report_path  # noqa 401
from birgitta.recipetest.coverage.report import cov_results  # noqa 401


@pytest.fixture(scope="session")
def spark_session():
    return spark.local_session()  # duration: about 3secs


context.set("TODAY", values.today())



"""Function for validating that two data sets are the same.
"""
from birgitta.dataframe import dataframe
from birgitta.dataframe import dfdiff


def datasets(spark_session, expected_ds, result_ds, project_key=None):
    """Validate that two data sets are the same.

    Args:
        spark_session (SparkSession): spark session used to load data frames.
        expected_ds (str): The expected data set to load.
        result_ds (str): The result data set to load.
        project_key (str): Used if data set in a separate dataiku project.
    """
    expected_df = dataframe.get(spark_session, expected_ds, prefix=project_key)
    result_df = dataframe.get(spark_session, result_ds, prefix=project_key)
    diff_ret = dfdiff.diff(expected_df, result_df)
    assert not diff_ret, "Dataframe diff: " + diff_ret
    print("Test successful")



"""Timing functions for recipetest.
"""
from datetime import datetime

__all__ = ['time', 'print_time', 'print_results']

first = datetime.now()
last = first
results = []


def time(description=""):
    """Add timing entry with description

    Args:
        description (str): Description for this entry.
    """
    global last, first, results
    entry = {
        "str": str,
        "last": last,
        "first": first,
        "now": datetime.now()
    }
    results.append(entry)
    last = entry["now"]


def print_time(entry):
    """Print the time for a single entry"""
    print("spent:",
          diffstr(entry["last"], entry["now"]),
          "total:",
          diffstr(entry["first"], entry["now"]),
          "now: ",
          entry["now"].strftime("%H:%M:%S.%f"),
          entry["str"])


def print_results(case=""):
    """print all the results for case"""
    global results
    print("=============== Time profile: %s ================\n" % (case))
    for entry in results:
        print_time(entry)


def diffstr(a, b):
    diff = b - a
    return str(diff)



"""Functions like `pause()`, `show()` and `die()` to enable debugging libs
from within a notebooks, thus also enabling lib sharing in
notebook work flows, helping to build production quality code.
"""
import pyspark


def die():
    """Manual breakpoint in notebook"""
    assert False


def pause():
    raw_input("Press the <ENTER> key to continue...")  # noqa: F821


def show(df, rows=3, truncate=False):
    if in_notebook():
        df.show(rows, truncate)


def mem_persist(df):
    if in_notebook():
        df.persist(pyspark.StorageLevel.MEMORY_ONLY)


def disk_persist(df):
    if in_notebook():
        df.persist(pyspark.StorageLevel.DISK_ONLY)


def in_notebook():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False



import dataiku
from birgitta import context
from birgitta.dataiku.dataset import manage as dataset_manage
from birgitta.dataiku.dataset.manage import schema
from birgitta.dataiku.recipe import manage as recipe_manage
from birgitta.recipetest import validate


def test_recipe(spark_session,
                scenario,
                src_project_key,
                src_recipe_key,
                testbench_project_key,
                test_params):
    # Trigger dataiku, not parquet
    context.set("BIRGITTA_DATASET_STORAGE", "DATAIKU")
    # Trigger dataiku, not parquet
    context.set("BIRGITTA_S3_BUCKET", "birgitta_s3_bucket")
    print('####################################################')
    print('Test recipe: %s (in project %s)' % (src_recipe_key,
                                               src_project_key))
    if src_project_key == testbench_project_key:
        raise ValueError('Cannot clone recipe to same project as src project')

    print('Clone dataset schemas')
    schemas = test_params['schemas']
    client = dataiku.api_client()
    cloned_input_datasets = schemas['inputs'].keys()
    cloned_input_datasets = clone_schemas(client,
                                          src_project_key,
                                          testbench_project_key,
                                          cloned_input_datasets,
                                          'Inline')
    cloned_output_datasets = schemas['outputs'].keys()
    cloned_output_datasets = clone_schemas(client,
                                           src_project_key,
                                           testbench_project_key,
                                           cloned_output_datasets,
                                           'HDFS')
    expected_output_datasets = create_expected_output_schemas(
        client,
        src_project_key,
        testbench_project_key,
        cloned_output_datasets
    )
    print('Clone recipe')
    recipe_manage.clone(client,
                        src_project_key,
                        src_recipe_key,
                        testbench_project_key,
                        test_name(src_recipe_key),
                        cloned_input_datasets,
                        cloned_output_datasets)

    test_cases = test_params['test_cases']
    for test_case in test_cases:
        print('Setup test case: ' + test_case['name'])
        print('Empty and fill datasets with fixtures')
        empty_and_fill_datasets(testbench_project_key,
                                cloned_input_datasets,
                                schemas['inputs'],
                                test_case['inputs'])
        empty_and_fill_datasets(testbench_project_key,
                                cloned_output_datasets,
                                schemas['outputs'],
                                False)  # empty dataset
        empty_and_fill_datasets(testbench_project_key,
                                expected_output_datasets,
                                expected_params(schemas['outputs']),
                                expected_params(test_case['outputs']))
        print('Run recipe')
        testbench_output_dataset_key = test_params['principal_output_dataset']
        scenario.build_dataset(dataset_name=testbench_output_dataset_key,
                               project_key=testbench_project_key)
        print('Validate output')
        for dataset_name in test_case['outputs']:
            print('Validate output dataset: %s' % (dataset_name))
            validate.datasets(spark_session,
                              dataset_name,
                              expected_name(dataset_name),
                              testbench_project_key)
            print('Successfully validated output dataset: %s' % (dataset_name))
    print('Delete testbench recipe TODO')
    print('Delete datasets TODO')
    print('Tests successful')


def test_name(recipe_name):
    return recipe_name + '_test'


def expected_name(dataset_name):
    return dataset_name + '_expected'


def delete_datasets(project, dataset_names):
    for dataset_name in dataset_names:
        dataset_manage.delete_if_exists(project, dataset_name)


def empty_and_fill_datasets(project_key,
                            dataset_names,
                            schemas,
                            row_sets=False):
    for dataset_name in dataset_names:
        rows = row_sets[dataset_name]['rows'] if row_sets else []
        dataset_manage.empty_and_fill(project_key,
                                      dataset_name,
                                      schemas[dataset_name],
                                      rows)


def clone_schemas(client,
                  src_project_key,
                  dst_project_key,
                  dataset_names,
                  output_type):
    datasets = []
    for dataset_name in dataset_names:
        datasets.append(dataset_name)
        schema.clone(client,
                     src_project_key,
                     dst_project_key,
                     dataset_name,
                     dataset_name,
                     output_type)

    return datasets


def create_expected_output_schemas(client,
                                   src_project_key,
                                   testbench_project_key,
                                   dataset_names):
    datasets = []
    for dataset_name in dataset_names:
        datasets.append(expected_name(dataset_name))
        schema.clone(client,
                     src_project_key,
                     testbench_project_key,
                     dataset_name,
                     expected_name(dataset_name),
                     'Inline')

    return datasets


def expected_params(set_params):
    ret = {}
    for dataset_name in set_params.keys():
        ret[expected_name(dataset_name)] = set_params[dataset_name]
    return ret



"""Module for test case fixturing functions.
"""
from birgitta import timing


__all__ = ['dataframes', 'write_fixtures']


def dataframes(fixtures, variant_name, spark_session):
    """Makes dataframes from fixtures.

    Args:
        fixtures (dict): Dict of fixtures
        variant_name (str): Name of fixture variant
        spark_session (SparkSesssion): Spark session used to create fixtures
        dataframe_source (DataframeSource): The source to write to, e.g. S3

    Returns:
       A dict of `dataset name => dataframe` pairs
    """
    # FUTURE: Try writing directly to parquet using pyarrow instead. Could be faster. # noqa 501
    ret = {}
    for fixture in fixtures:
        ret[fixture.dataset.name] = fixture.df(variant_name,
                                               spark_session)
    return ret


def write_fixtures(fixtures, variant_name, spark_session, dataframe_source):
    """Write fixtures to storage.

    Args:
        fixtures (dict): Dict of fixtures
        variant_name (str): Name of fixture variant
        spark_session (SparkSession): Spark session used to create fixtures
        dataframe_source (DataframeSource): The source to write to, e.g. FS
    """
    timing.time("write_fixtures start")
    dfs = dataframes(fixtures, variant_name, spark_session)
    for ds_name in dfs.keys():
        dataframe_source.write(dfs[ds_name], ds_name)
    timing.time("write_fixtures end")



"""Field / feature catalog class. This catalog is per project/domain.
It holds a list of fields with description and example values.
It does not hold field types. The field types themselves are defined
in specific data set schemas, for clarity and more intuitive schema APIs.
Additionally, by decoupling data type from field description and example_val,
The same example value e.gg. 2017-01-01 can be interpreted as date, datetime
or string, depending on the schema needs.

The purpose of the catalog is:

1. generate example fixtures for recipe tests in an efficient manner
2. reuse default values across data set fixtures for simplicity
3. promote reuse of field names across data sets
4. document all fields and eventually enable automatic publishing of these
comments in views and reports
"""
from birgitta.schema.fixtures import ExampleVal


__all__ = ['Catalog']


class Catalog:
    def __init__(self):
        """Creates an empty catalog.
        """
        self.field_confs = {}
        self.secondary_catalog = None

    def set_secondary(self, catalog):
        """Set secondary catalog to fall back on for missing fields.
        Args:
            secondary_catalog (Catalog): fall back Catalog
        """
        self.secondary_catalog = catalog

    def copy(self):
        cat = Catalog()
        cat.field_confs = self.field_confs
        cat.set_secondary(self.secondary_catalog)
        return cat

    def add_field(self,
                  name,
                  *,
                  example,
                  description):
        self.field_confs[name] = {
            "example": example,
            "description": description
        }

    def example_val(self,
                    field):
        """Get the example val for a field.

        Args:
            field (str): Field name.
            override_conf (dict): Overrides catalog fixture conf.

        Returns:
            example value
        """
        field_conf = self.get_field_conf(field)
        return ExampleVal(field_conf['example'])

    def get_field_conf(self, field):
        if field in self.field_confs:
            return self.field_confs[field]
        # If not in primary field_confs, look in secondary catalog
        if not self.secondary_catalog:
            raise ValueError("Missing field %s" % (field))
        return self.secondary_catalog.get_field_conf(field)



"""library for testing custom python libraries on dataiku

## Examples

How to run all test_*.py files within pytest in a dataiku project or
system library called foo:

from birgitta.dataiku.libtest import libtest

import foo

libtest.test(foo)
"""
import inspect
import os

import pytest


def test(lib):
    test_path = os.path.dirname(inspect.getfile(lib))
    ret = pytest.main(["-x", test_path + "/tests", "-vv"])
    success = ret == 0
    if not success:
        print(f"Pytest failed on library {test_path}")
    return success



"""The Dataiku dataframesource class.
"""
import sys

from birgitta.dataframesource import DataframeSourceBase
from pyspark.sql.utils import AnalysisException


__all__ = ['LocalSource']


class LocalSource(DataframeSourceBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if 'dataset_dir' in kwargs:
            self.dataset_dir = kwargs['dataset_dir']
        else:
            self.dataset_dir = None

    def load(self, spark_session, dataset_name, prefix, **kwargs):
        dataset_dir = prefix  # Interpret prefix as dataset_dir
        if not dataset_dir:
            dataset_dir = self.dataset_dir
        try:
            read = spark_session.read
            schema = kwargs.get('schema')
            if schema:
                read = read.schema(schema.to_spark())
            return read.parquet(
                "%s/%s/" % (dataset_dir, dataset_name))
        except AnalysisException as e:
            if "Path does not exist" in str(e):
                print("e.desc:", repr(e.desc))
                err_msg = (
                    "Missing dataset: %s, " +
                    "Is the dataset name correct and equal to the dataset table_name for the fixture? " +  # noqa 501
                    "Error: %s") % (dataset_name, e.desc)
                print("err_msg:", repr(err_msg))
                raise type(e)(err_msg, sys.exc_info()[2])
            else:
                raise e

    def write(self, df, dataset_name, dataset_dir=None, **kwargs):
        if not dataset_dir:
            dataset_dir = self.dataset_dir
        dest_path = "%s/%s/" % (dataset_dir, dataset_name)
        # Use coalesce to write a single file, and thus preserve row order
        df.coalesce(1).write.format("parquet").mode("append").save(dest_path)



import datetime
from datetime import timedelta

from pyspark.sql import functions as F
from pyspark.sql import types as T

__all__ = ['group_by_day']


def group_by_day(df,
                 *,
                 spark_session,
                 group_keys,
                 from_date_field,
                 to_date_field,
                 join_partitions=100,
                 count_partitions=3,
                 min_cutoff_date=None,
                 max_cutoff_date=datetime.date.today(),
                 drop_date_fields=True):
    """Count entries per day grouped by the keys in group_keys.

    Args:
        df (DataFrame): Incoming spark dataframe.

    Kwargs:
        spark_session (SparkSession): Spark session used to build date
        dataframe.
        group_keys (list): List of strings with keys to group by.
        from_date_field (str): From data field name.
        to_date_field (str): To data field name.
        join_partitions (int): Number of partitions for the join with dates.
        count_partitions (int): Number of partitions after the count.
        min_cutoff_date (datetime.date): Drop all data before this date.
        max_cutoff_date (datetime.date): Drop all data after this date,
        can often be today.
        drop_date_fields (bool): If True drop original from_date and to_date
        fields from result.
    Returns:
       Spark DataFrame with grouped counts per key combination per day.

    Example input:
    DataFrame
    +-------------+---------+-----------+-----------+---------+--------+
    |customer_id  |prod_code|shop_code  |channel    |from_date| to_date|
    +-------------+---------+-----------+-----------+---------+--------+
    |         1234|      ABC|      SHOP1|          A| 20160606|20160610|
    |         3456|      ABC|      SHOP1|          A| 20160607|20160611|
    +-------------+---------+-----------+-----------+---------+--------+
    group_keys=['prod_code', 'shop_code', 'channel']
    from_date_field='from_date'
    to_date_field='to_date'
    drop_date_fields=True

    Example output:
    DataFrame
    +--------+-----------+------------+---------+-----+
    | datenum|channel    |shop_code   |prod_code|count|
    +--------+-----------+------------+---------+-----+
    |20160606|          A|       SHOP1|      ABC|    1|
    |20160607|          A|       SHOP1|      ABC|    2|
    +--------+-----------+------------+---------+-----+
    """
    df_with_dates = df\
        .withColumn('date_from', F.col(from_date_field))\
        .withColumn('date_to', F.col(to_date_field))

    dates_range = df_with_dates.agg(
        F.min(F.col("date_from")),
        F.max(F.col("date_to"))
    ).collect()

    all_dates = all_dates_between(dates_range[0][0],
                                  dates_range[0][1],
                                  min_cutoff_date,
                                  max_cutoff_date)
    dates_df = spark_session.createDataFrame(all_dates, T.DateType())
    between = dates_df['value'].between(df_with_dates['date_from'],
                                        df_with_dates['date_to'])
    result = df_with_dates.repartition(join_partitions)\
                          .join(dates_df, on=between)

    datenum_type = F.date_format(F.col('datenum'), 'YYYYMMdd').cast("int")
    result = result\
        .withColumn('datenum', result['value'])\
        .withColumn('datenum', datenum_type)\
        .drop('value')\
        .drop('date_from')\
        .drop('date_to')

    if drop_date_fields:
        result = result.drop(from_date_field).drop(to_date_field)

    new_group_keys = group_keys[:]
    new_group_keys.append('datenum')
    return result.groupBy(new_group_keys)\
                 .count().repartition(count_partitions)


def all_dates_between(ds_min_date,
                      ds_max_date,
                      min_cutoff_date,
                      max_cutoff_date):
    ds_min_date = ensure_date(ds_min_date)
    ds_max_date = ensure_date(ds_max_date)
    if min_cutoff_date:
        min_date = max(ds_min_date, min_cutoff_date)
    else:
        min_date = ds_min_date
    if max_cutoff_date:
        max_date = min(ds_max_date, max_cutoff_date)
    else:
        max_date = ds_max_date

    delta = max_date - min_date
    all_dates = []
    for i in range(delta.days + 1):
        day = min_date + timedelta(days=i)
        all_dates.append(day)
    return all_dates


def ensure_date(dt):
    if type(dt) == datetime.datetime:
        return dt.date()
    return dt



"""Functionality for running fixture based test cases.
"""
import os
from functools import partial

import pytest
from birgitta import timing
from birgitta.dataframesource.sources.localsource import LocalSource
from birgitta.recipe import runner
from birgitta.recipetest import localtest
from birgitta.recipetest.coverage import report
from birgitta.recipetest.localtest import fixturing, assertion, script_prepend  # noqa F401
from birgitta.recipetest.coverage.report import cov_report, dbg_counts, cov_results  # noqa F401
from birgitta.recipetest.coverage.transform import prepare

__all__ = ['recipe_path', 'run_case_partial']


@pytest.fixture("function")  # noqa F401
def run_case_partial(tmpdir,
                     cov_report,  # noqa F811
                     cov_results,  # noqa F811
                     spark_session,
                     request):
    """Create a run_case partial to enable the running fixture
    based cases, referring only to the fixture name. Example:

    ```
    def run_case(run_case_partial):   # noqa F811
    return run_case_partial([[ds_contracts, fx_contracts]],
                            [[ds_filtered_contracts,
                              fx_filtered_contracts]],
                            recipe_path(__file__))


    def test_default(run_case):
        run_case("default")
    ```

    Args:
        input_datasets (list): A list with an entry of for each input dataset,
        each entry is a list of the corresponding dataset and fixture module.
        expected_datasets (list): A list with an entry of for each output dataset,
        each entry is a list of the corresponding dataset and fixture module.
        recipe_path (str): File path of the spark recipe.

    Returns:
        the partial with its required resources.
    """
    test_case = request.node.name
    return partial(transposed_partial,
                   tmpdir,
                   cov_report,
                   cov_results,
                   test_case,
                   spark_session)


def recipe_path(file_path, prefix="compute_"):
    """Derive the recipe path given the test path.

    Args:
        file_path (str): Path to the test_ file.
        prefix (str): Prefix of the spark recipe, defaults to '_compute'.

    Returns:
        File path of the recipe.
    """
    dir_path = os.path.dirname(os.path.abspath(file_path))
    filename = os.path.basename(file_path).replace("test_", prefix)
    return F"{dir_path}/../../recipes/{filename}" # noqa F501


def transposed_partial(*argv):
    """Utility function to transpose list of partial args"""
    args = list(argv)
    args.insert(0, run_case)
    return partial(*args)


def run_case(tmpdir,
             cov_report_file,
             cov_results,  # noqa F811
             test_case,
             spark_session,
             in_fixtures,
             out_fixtures,
             recipe_path,
             fixture_name):
    """Run a test case.

    Does the following:

    * Reporting and timing.
    * Setup input and result fixtures.
    * Run the spark script in recipe_path.
    * Asserts the outputs.
    * Collects and presents report.

    Returns:
        None
    """
    timing.time("run_case_fn start: %s" % (fixture_name))
    tdir = tmpdir.strpath
    dataframe_source = LocalSource(dataset_dir=tdir)
    fixture_name = test_case[5:]  # Lose 'test_' prefix
    fixturing.write_fixtures(in_fixtures,
                             fixture_name,
                             spark_session,
                             dataframe_source)
    expected_dfs = fixturing.dataframes(out_fixtures,
                                        fixture_name,
                                        spark_session)
    localtest.process_recipe(recipe_path,
                             tdir,
                             dataframe_source,
                             cov_report_file,
                             test_case,
                             spark_session)
    timing.time("runcase_fn run_script done: %s" % (fixture_name))
    assertion.assert_outputs(expected_dfs,
                             dataframe_source,
                             spark_session)
    report.collect(cov_report_file, test_case, cov_results)
    timing.time("run_case_fn end: %s" % (fixture_name))
    timing.print_results(test_case)


def process_recipe(path,
                   tdir,
                   dataframe_source,
                   cov_report_file,
                   test_case,
                   spark_session):
    """Prepare the recipe and trigger the recipe execution.
    """
    with open(path) as f:
        code = f.read()
    code_w_reporting = prepare(code)
    cov_dict = {}
    cov_dict["cov_report_file"] = cov_report_file
    cov_dict["test_case"] = test_case

    globals_dict = {
        "BIRGITTA_SPARK_SESSION_TYPE": "LOCAL",
        "BIRGITTA_TEST_COVERAGE": cov_dict,
        "BIRGITTA_DBG_COUNTS": dbg_counts()
    }
    full_code = script_prepend.code(tdir) + code_w_reporting
    dump_test_recipe(test_case, tdir, full_code)
    timing.time("execute_recipe before exec")
    runner.exec_code(full_code, globals_dict)
    timing.time("execute_recipe after exec")


def dump_test_recipe(test_case, tmpdir, code):
    dump_path = tmpdir + "/" + test_case + ".py"
    print("\nTest recipe python file:\n", repr(dump_path), "\n")
    with open(dump_path, "w") as f:
        f.write(code)



"""Utility functions for manipulating dataiku datasets.
Used by dataiku recipetest."""
import dataiku
from pandas import DataFrame


def delete_if_exists(project, name):
    for dataset_meta in project.list_datasets():
        if name == dataset_meta['name']:
            dataset_to_delete = project.get_dataset(name)
            dataset_to_delete.delete()
            break


def empty_and_fill(project_key, dataset_name, schema, rows):
    dst_ds = dataiku.Dataset(dataset_name, project_key)
    fill_df = DataFrame(rows)
    dst_ds.write_with_schema(fill_df, True)  # True => dropAndCreate



"""Evaluate coverage of all transformations in notebook,
ensuring that all of them have handled at least one record, to
promote test coverage of all logic.
"""

import json
import os

import pytest
from birgitta import context
from birgitta import timing

__all__ = [
    'log_transform',
    'collect',
    'cov_report_path',
    'cov_report',
    'cov_results',
    'dbg_counts',
    'transform_coverage'
]


@pytest.fixture(scope="function")
def transform_coverage(request, cov_report_path, cov_results):
    """transform_coverage fixture enables the verification of
    complete transform coverage accumulated for all test cases
    for a recipe. transform_coverage test case is added this way:

    def test_coverage(transform_coverage):  # noqa 811
        print("Validate coverage")

    No specification is needed.
    """
    assign_counts = {}
    raw_report = cov_results['raw']
    for test_case in raw_report.keys():
        get_assign_counts(raw_report[test_case], assign_counts)
    print("\nData frame assign counts:\n")
    assign_counts_print(assign_counts)

    for assign_key in assign_counts.keys():
        record_count = assign_counts[assign_key]["record_count"]
        assert record_count > 0, "Assign has zero count: " + assign_key


def log_transform(var, line_no, line_str):
    """Helper function to log a transformation.
    localtest will insert log_transform() calls after assignments.
    The call contains the line_no and the assignement code.
    The purpose is to log the transformation, including the
    dataframe count if it is a dataframe. The purpose is to
    Verify that more than zero rows pass through each transform.

    Args:
        var (str): The variable being logged
        line_no (int): The line number of the original code line
        line_str (str): The original code line
    """
    report_file = context.get("BIRGITTA_TEST_COVERAGE")["cov_report_file"]
    test_case = context.get("BIRGITTA_TEST_COVERAGE")["test_case"]
    type_name = type(var).__name__
    metrics = {"var_type": type_name}
    if type_name == "DataFrame":
        metrics["count"] = var.count()
    log_entry(test_case, line_no, line_str, report_file, metrics)


def collect(cov_report, test_case, cov_results):
    """Collect all transform reports for test_case and
    print case report.

    Args:
        cov_report (str): report file
        test_case (str): test case being collected
        cov_results (dict): dict holding the results format
        for all test cases. Results for test_case are added.
    """
    line_executions = []
    with open(cov_report) as f:
        report_lines = f.readlines()
        for report_line in report_lines:
            line_executions.append(json.loads(report_line))
        cov_results["raw"][test_case] = line_executions
    print_case_report(test_case, line_executions)


@pytest.fixture(scope="function")
def cov_report(cov_report_path):
    """Ensure coverage report path is setup when running tests."""
    if not os.path.exists(cov_report_path):
        cov_report_path.write("")
    return cov_report_path.strpath


@pytest.fixture(scope="function")
def cov_report_path(tmpdir):
    """Force coverage.log to be present when running tests."""
    return tmpdir.mkdir("coverage").join("coverage.log")


@pytest.fixture("module")  # noqa 401
def cov_results():
    """ Fixture holding transform coverage results from all test cases
    """
    return {
        "raw": {}
    }


def dbg_counts():
    """Check if env var BIRGITTA_DBG_COUNTS is set.
    If true, print out transformation counts when running test."""
    return os.environ.get('BIRGITTA_DBG_COUNTS') is not None


def log_entry(test_case, line_no, line_str, report_file, metrics):
    """Log a report entry for line of the recipe.
    """
    timing.time(line_str)
    if dbg_counts() and (metrics['var_type'] == 'DataFrame'):
        print("l:", line_no, repr(line_str), "count:", metrics['count'])
    with open(report_file, 'a') as f:
        json_dict = {
            "test_case": test_case,
            "line_no": line_no,
            "line_str": line_str,
            "metrics": metrics,
        }
        f.write(json.dumps(json_dict))
        f.write("\n")


def print_case_report(test_case, line_executions):
    """Print count report for the current test_case
    """
    print("\nData frame assign counts for test case " + test_case + ":\n")
    assign_counts = get_assign_counts(line_executions)
    assign_counts_print(assign_counts)


def assign_counts_print(assign_counts):
    """Print assign counts for each line.
    """
    for assign_key in assign_counts:
        assign_count = assign_counts[assign_key]
        print(assign_count["record_count"], "records,",
              assign_count["executions"], "executions,",
              "line:", assign_key)
    print("\n")


def get_assign_counts(line_executions, assign_counts={}):
    """Gather the assign counts for each line.
    """
    for line_execution in line_executions:
        metrics = line_execution['metrics']
        if metrics['var_type'] != "DataFrame":
            continue
        assign_key = "%d:%s" % (line_execution['line_no'],
                                line_execution['line_str'])
        if assign_key not in assign_counts:
            assign_counts[assign_key] = {
                "executions": 0,
                "record_count": 0
            }
        assign_counts[assign_key]['executions'] += 1
        assign_counts[assign_key]['record_count'] += metrics['count']
    return assign_counts



from dataikuapi import CodeRecipeCreator
from dataikuapi.dss.recipe import DSSRecipe


def delete_if_exists(project, key):
    for recipe_meta in project.list_recipes():
        # pprint(recipe_meta)
        if key == recipe_meta['name']:
            recipe_to_delete = project.get_recipe(key)
            recipe_to_delete.delete()
            break


def clone(client,
          src_project_key,
          src_recipe_key,
          testbench_project_key,
          testbench_recipe_key,
          input_dataset_names,
          output_dataset_names):
    # src_dataset = src_project.get_dataset(src_fixtures_dataset_key)
    testbench_project = client.get_project(testbench_project_key)
    delete_if_exists(testbench_project, testbench_recipe_key)

    src_recipe = DSSRecipe(client, src_project_key, src_recipe_key)
    definition_and_payload = src_recipe.get_definition_and_payload()
    # Assume pyspark recipe, FUTURE support other types
    recipe_type = "pyspark"
    builder = CodeRecipeCreator(testbench_recipe_key,
                                recipe_type,
                                testbench_project)
    # source code
    builder = builder.with_script(definition_and_payload.get_payload())

    # Add inputs and outputs
    for dataset_name in input_dataset_names:
        builder = builder.with_input(dataset_name)
    for dataset_name in output_dataset_names:
        builder = builder.with_output(dataset_name)
    recipe = builder.build()

    return recipe



"""RowConfsVariant - class for fixtures with row_confs.
Row confs derive fixture values from catalogs and schemas.
"""

from .variant_base import VariantBase
from ..row_conf import RowConf
from ... import fixtures

__all__ = ['RowConfsVariant']


class RowConfsVariant(VariantBase):
    def __init__(self, collection):
        for entry in collection:
            if not (type(entry) == RowConf):
                raise ValueError(
                    "Entry (%s) must be RowConf" % repr(entry))
        self.row_confs = collection

    def df(self, spark, dataset):
        return fixtures.df(spark,
                           dataset.schema,
                           self.row_confs)



"""VariantBase, the base class for fixture variants.
This base class enables us to create fixtures in different ways.
"""

from abc import ABCMeta, abstractmethod

__all__ = ['VariantBase']


class VariantBase(metaclass=ABCMeta):
    def __init__(self, collection):
        """Abstract class for fixture variants.

        Args:
            collection (list of lists): The rows or row_confs,
            depending on class.
        """
        pass

    @abstractmethod
    def df(self, spark):
        """Method for creating the fixture data frame.

        Args:
            spark (SparkContext): Spark context to generate data frame

        Returns:
            A data frame of the fixture."""
        return NotImplemented
